"""Utilities used for testing"""

import json
import struct
import wave
from pathlib import Path
from typing import Dict, List


def create_word_files(
    audio_path: Path, files: List[str], touch_only: bool = False
) -> List[Path]:
    """Create word files for testing

    Args:
        audio_path (Path): Path to create word files in
        word_strings (List[str]): Words to create files for

    Returns:
        List[Path]: List of paths to created word files
    """
    word_files = []
    for file in files:
        filepath = audio_path.joinpath(file)
        word_files.append(filepath)

        if touch_only:
            filepath.touch()
            continue
        savefile = wave.open(str(filepath), "w")

        # channels, datasize (16 bit), sample rate, number of samples
        savefile.setparams((1, 2, 11025, 500, "NONE", "Uncompressed"))
        savefile.writeframes(struct.pack("h", 1))
        savefile.close()
    return word_files


def create_voice_files(
    base_path: Path,
    files: List[str],
    voice_name: str = "voice",
    touch_only: bool = False,
) -> Path:
    """Create voice files for testing

    Args:
        base_path (Path): Path to create voice folder at
        files (List[str]): Files to add to voice folder
        voice_name (str, optional): Voice name. Defaults to 'voice'.
        touch_only (bool, optional): Only touch voice files, don't use actual audio. Defaults to False.

    Returns:
        Path: Path to voice folder
    """
    audio_path = base_path.joinpath(voice_name)
    audio_path.mkdir()

    create_word_files(audio_path=audio_path, files=files, touch_only=touch_only)

    return audio_path


def create_category_files(
    base_path: Path, category_files: Dict[str, List[str]], voice_name: str = "voice"
) -> Path:
    """Create category files for testing

    Args:
        base_path (Path): Path to create category files in
        category_files (Dict[str, List[str]]): Files to create
        voice_name (str, optional): Voice name. Defaults to 'voice'.

    Returns:
        Path: Path to category files
    """
    audio_path = base_path.joinpath(voice_name)
    audio_path.mkdir()

    for category, files in category_files.items():
        create_voice_files(base_path=audio_path, files=files, voice_name=category)

    return audio_path


def create_info(info: dict, audio_path: Path) -> None:
    """Create info file for testing

    Args:
        info (dict): Info contents
        audio_path (Path): Path to create info file in
    """
    info_dir = Path(audio_path).joinpath("info/")

    info_dir.mkdir(parents=True, exist_ok=True)

    info_file = info_dir.joinpath("info.json")

    with open(info_file, "w", encoding="UTF-8") as file:
        json.dump(info, file)



"""Manages multiple voices"""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Union, Type
from types import TracebackType

import sqlalchemy

from hlvox.multi_voice import MultiVoice
from hlvox.single_voice import SingleVoice
from hlvox.voice import Voice

log = logging.getLogger(__name__)


class DuplicateVoice(Exception):
    """Raised when duplicate voices are found"""


@dataclass
class RemoteDatabaseInfo:
    """Database connection information if using remote database (sql)"""

    name: str
    url: str
    port: int
    username: str
    password: str


@dataclass
class LocalDatabaseInfo:
    """Database info if using local file-based database (sqlite)"""

    base_path: Path


class Manager:
    """Manages multiple voices"""

    def __init__(
        self,
        voices_path: Union[Path, str],
        database_info: Optional[Union[LocalDatabaseInfo, RemoteDatabaseInfo]],
    ):
        self.voices_path = Path(voices_path)
        self._database_info = database_info
        self.voices: Dict[str, Union[MultiVoice, SingleVoice]] = {}

        self.scan_voices()

    def __enter__(self) -> "Manager":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.exit()

    def _create_db_path(self, databases_path: Path, voice_name: str) -> Path:
        db_path = databases_path / voice_name
        db_path.mkdir(parents=True, exist_ok=True)
        return db_path

    def _create_db(
        self,
        database_info: Union[LocalDatabaseInfo, RemoteDatabaseInfo],
        voice_name: str,
    ) -> sqlalchemy.engine.Engine:
        dbi = database_info
        if isinstance(dbi, LocalDatabaseInfo):
            path = self._create_db_path(
                databases_path=dbi.base_path, voice_name=voice_name
            )
            return sqlalchemy.create_engine(f"sqlite:///{path}/db.sqlite")
        return sqlalchemy.create_engine(
            f"postgresql+psycopg2://{dbi.username}:{dbi.password}@{dbi.url}:{dbi.port}/{dbi.name}",
            pool_pre_ping=True,
        )

    def scan_voices(self) -> None:
        """Scan for voices in the voices_path, updating the voices dictionary"""
        # Clean up existing voices and their database connections, if any
        for voice in self.voices.values():
            voice.exit()
        self.voices.clear()

        # Load new voices
        # This strangeness is to make mypy happy. There is probably a cleaner way to do it.
        single_voices = self._load_voices(self.voices_path)
        multi_voice = self._create_multi_voice(single_voices)
        self.voices.update(single_voices)
        self.voices["multi"] = multi_voice

    def _load_voices(self, path: Path) -> Dict[str, SingleVoice]:
        voices = {}
        voice_folders = list(x for x in path.iterdir() if x.is_dir())
        logging.info("Found %d voice folders", len(voice_folders))
        for voice_folder in voice_folders:
            voice_name = voice_folder.name.lower()
            logging.info("Loading voice %s", voice_name)
            if voice_name in voices:
                raise DuplicateVoice("Duplicate voice name found")

            database = None
            if self._database_info is not None:
                database = self._create_db(
                    database_info=self._database_info, voice_name=voice_name
                )
            new_voice = SingleVoice(
                name=voice_name, path=voice_folder, database=database
            )
            voices[new_voice.name] = new_voice
        logging.info("Loaded %d voices", len(voices))
        return voices

    def _create_multi_voice(self, voices: Dict[str, SingleVoice]) -> MultiVoice:
        database = None
        if self._database_info is not None:
            database = self._create_db(
                database_info=self._database_info, voice_name="multi"
            )
        return MultiVoice(
            voices=voices,
            database=database,
        )

    def get_voice_names(self) -> List[str]:
        """Gets names of available voices

        Returns:
            list -- list of voice name strings
        """

        voice_names = list(self.voices.keys())
        voice_names.sort()
        return voice_names

    def get_voice(self, name: str) -> Optional[Voice]:
        """Get voice of requested name

        Args:
            name ({string}): name of voice to get

        Returns:
            {voxvoice}: requested voice
        """
        if name in self.voices:
            return self.voices[name]
        return None

    def exit(self) -> None:
        """Exit all loaded voices"""
        for voice in self.voices.values():
            voice.exit()



"""Test voice manager"""

# pylint: disable=missing-function-docstring,disable=missing-class-docstring
# pylint: disable=too-few-public-methods
from typing import Generator
import sys
from pathlib import Path

import pytest

import hlvox
import hlvox.manager

from . import utils as th

test_files = {"v1": ["hello.wav"], "v2": ["hello.wav"], "v3": ["hello.wav"]}


@pytest.fixture(name="manager")
def manager_fixture(tmp_path: Path) -> Generator[hlvox.manager.Manager, None, None]:
    voices_dir = tmp_path.joinpath("voices")
    voices_dir.mkdir()
    for voice, words in test_files.items():
        th.create_voice_files(voices_dir, words, voice)
    manager = hlvox.manager.Manager(voices_path=voices_dir, database_info=None)
    yield manager
    manager.exit()


class TestVoiceImport:
    def test_voice_list(self, manager: hlvox.manager.Manager) -> None:
        expected_voices = list(test_files.keys())
        expected_voices.append("multi")
        expected_voices.sort()
        assert manager.get_voice_names() == expected_voices

        # Add a new voice
        new_voice_name = "v4"
        new_words = ["hello.wav"]
        new_voice_path = th.create_voice_files(
            manager.voices_path, new_words, new_voice_name
        )
        manager.scan_voices()
        new_expected_voices = expected_voices.copy()
        new_expected_voices.append(new_voice_name)
        new_expected_voices.sort()
        assert manager.get_voice_names() == new_expected_voices

        # Remove a voice
        for file in new_voice_path.glob("*"):
            file.unlink()
        new_voice_path.rmdir()
        manager.scan_voices()
        assert manager.get_voice_names() == expected_voices


class TestGetVoice:
    def test_get_voice(self, manager: hlvox.manager.Manager) -> None:
        voice = manager.get_voice("v1")

        assert voice is not None
        assert voice.name == "v1"

    def test_get_wrong_voice(self, manager: hlvox.manager.Manager) -> None:
        voice = manager.get_voice("nope")

        assert not voice


class TestDuplicatevoice:
    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="Windows is not case sensitive so duplicates cant exist",
    )
    def test_case_diff(self, tmp_path: Path) -> None:
        voices_dir = tmp_path.joinpath("voices")
        voices_dir.mkdir()
        th.create_voice_files(voices_dir, ["hello.wav"], "AHH")
        th.create_voice_files(voices_dir, ["hello.wav"], "ahh")
        th.create_voice_files(voices_dir, ["hello.wav"], "aHH")

        with pytest.raises(hlvox.manager.DuplicateVoice):
            hlvox.manager.Manager(voices_path=voices_dir, database_info=None)



# pylint: disable=too-many-lines
"""Test voice"""
# NOTE: since these tests use sqlite as their DB backend, indexes for results
# start at 1. This may differ if using postgresql in actual use.

# pylint: disable=missing-function-docstring,disable=missing-class-docstring
# pylint: disable=too-few-public-methods
# TODO: This isn't great and implies some re-architecting is needed
# pylint: disable=protected-access
import sys
from pathlib import Path
from typing import Generator, List, Optional

import pytest
import sqlalchemy

import hlvox
from hlvox.single_voice import SingleVoice
from hlvox.multi_voice import MultiVoice
from hlvox.voice import (
    NoOpModifier1,
    SentenceIdNotFound,
    SpeedChangeModifier,
    Word,
)

from . import utils as th

# Stand-in files for testing
normal_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav"]
normal_files_words = [Word(word=word[:-4], voice="test") for word in normal_files]
inconst_format_files = ["hello.mp3", "my.wav", "name", "is.wav", "vox.mp4"]
no_format_files = ["imatextfile", "whatami"]
alph_files = ["a.wav", "b.wav", "c.wav"]
alph_files_words = [Word(word=word[:-4], voice="test") for word in alph_files]
category_files = {
    "bad": ["stop.wav", "bad.wav", "no.wav"],
    "good": ["go.wav", "yes.wav"],
}


def words_from_strings(
    word_strings: List[str], voice: Optional[str] = None, punctuation: bool = False
) -> List[Word]:
    return [
        Word(word=word, voice=voice, is_punctuation=punctuation)
        for word in word_strings
    ]


class TestWord:
    def test_sorting(self) -> None:
        words = ["hi", "a", "zebra", "bit", "me"]
        word_classes = words_from_strings(words)
        word_classes.sort()

        words_sorted = words.copy()
        words_sorted.sort()
        word_classes_sorted = words_from_strings(words_sorted)

        assert word_classes == word_classes_sorted

    def test_list(self) -> None:
        # This is a really roundabout way to test equality and such,
        # but its something I ran into so _shrug_
        word1 = Word(word="hello")
        word2 = Word(word="hello")

        word_list = [word1]
        assert word2 in word_list

    def test_as_str(self) -> None:
        simple = Word(word="hello")
        assert simple.as_str() == "hello"
        assert simple.as_str(with_voice=True) == "hello"

        voice = Word(word="hello", voice="voice")
        assert voice.as_str() == "hello"
        assert voice.as_str(with_voice=True) == "voice:hello"

        modifier = Word(word="hello", voice="voice", modifiers=[NoOpModifier1(1.0)])
        assert modifier.as_str() == "hello|z1.0"
        assert modifier.as_str(with_voice=True) == "voice:hello|z1.0"

        modifiers = Word(
            word="hello",
            voice="voice",
            modifiers=[
                NoOpModifier1(parameter=2),
                SpeedChangeModifier(speed_parameter=0.1),
            ],
        )
        assert modifiers.as_str() == "hello|s0.1,z2.0"
        assert modifiers.as_str(with_voice=True) == "voice:hello|s0.1,z2.0"


class TestFileHandling:
    def test_empty_files(self, tmp_path: Path) -> None:
        voice_dir = th.create_voice_files(tmp_path, [])
        with pytest.raises(hlvox.voice.NoWordsFound):
            hlvox.single_voice.SingleVoice(name="test", path=voice_dir, database=None)


class TestDictContents:
    def test_basic_dict(self, tmp_path: Path) -> None:
        voice_dir = th.create_voice_files(tmp_path, normal_files)

        expected_names = normal_files_words
        expected_names.sort()

        with hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=None
        ) as unit:
            assert expected_names == unit.words

    def test_alphab(self, tmp_path: Path) -> None:
        voice_dir = th.create_voice_files(tmp_path, alph_files)

        expected_names = alph_files_words
        expected_names.sort()

        with hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=None
        ) as unit:
            assert expected_names == unit.words

    def test_caps(self, tmp_path: Path) -> None:
        word_strings = [
            "Cap.wav",
            "Cappa.wav",
        ]
        voice_dir = th.create_voice_files(tmp_path, word_strings)

        expected_words = [
            Word(word=word.lower()[:-4], voice="test") for word in word_strings
        ]
        expected_words.sort()

        with hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=None
        ) as unit:
            assert expected_words == unit.words

    def test_categories(self, tmp_path: Path) -> None:
        voice_dir = th.create_category_files(tmp_path, category_files)
        expected_names = [
            item[:-4] for sublist in category_files.values() for item in sublist
        ]
        expected_words = words_from_strings(expected_names, voice="test")
        expected_words.sort()
        with hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=None
        ) as unit:
            assert expected_words == unit.words

            for category, words in category_files.items():
                category_words = [word[:-4] for word in words]
                category_words.sort()
                assert unit.categories[category] == category_words

    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="Windows is not case sensitive so duplicates cant exist",
    )
    def test_duplicates(self, tmp_path: Path) -> None:
        words = [
            "Cap.wav",
            "cAP.wav",
        ]
        voice_dir = th.create_voice_files(tmp_path, words, touch_only=True)

        with pytest.raises(hlvox.voice.DuplicateWords):
            hlvox.single_voice.SingleVoice(name="test", path=voice_dir, database=None)


class TestDatabase:
    def test_reuse(self, tmp_path: Path) -> None:
        voice_dir = th.create_voice_files(tmp_path, normal_files)
        database = sqlalchemy.create_engine(f"sqlite:///{tmp_path}/db.sqlite")
        voice = hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=database
        )

        voice.generate_audio(sentence="hello hello")
        voice.exit()

        voice = hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=database
        )
        assert voice.get_generated_sentences() == ["hello hello"]


class TestScanFiles:
    def test_scan_files(self, tmp_path: Path) -> None:
        voice_dir = th.create_voice_files(tmp_path, normal_files)
        voice = hlvox.single_voice.SingleVoice(
            name="test", path=voice_dir, database=None
        )
        assert voice.words == normal_files_words

        # Add new files
        new_word_strings = ["new", "words"]
        new_words = [Word(word=word, voice="test") for word in new_word_strings]
        new_word_files = th.create_word_files(voice_dir, new_word_strings)

        expected_words = normal_files_words + new_words
        expected_words.sort()
        voice.scan_files()
        assert voice.words == expected_words

        # Remove a file
        new_word_files[0].unlink()
        voice.scan_files()
        expected_words = normal_files_words + new_words[1:]
        expected_words.sort()
        assert voice.words == expected_words


@pytest.fixture(name="voice")
def voice_fixture(tmp_path: Path) -> Generator[SingleVoice, None, None]:
    voice_dir = th.create_voice_files(tmp_path, normal_files)
    database = sqlalchemy.create_engine(f"sqlite:///{tmp_path}/db.sqlite")
    voice = hlvox.single_voice.SingleVoice(
        name="test", path=voice_dir, database=database
    )
    yield voice
    voice.exit()


class TestSayableUnsayable:
    def test_empty_sent(self, voice: SingleVoice) -> None:
        ret_say, ret_unsay = voice.get_sayable_unsayable([])

        assert ret_say == []
        assert ret_unsay == []

    def test_simple_sent(self, voice: SingleVoice) -> None:
        words = words_from_strings(["hello"], voice="test")
        ret_say, ret_unsay = voice.get_sayable_unsayable(words)

        assert ret_say == [Word(word="hello", voice="test")]
        assert ret_unsay == []

    def test_duplicates(self, voice: SingleVoice) -> None:
        words = words_from_strings(
            ["hello", "hello", "world", "world", "duplicates!", "duplicates"],
            voice="test",
        )
        words.extend(
            words_from_strings([",", ",", ".", "."], voice="test", punctuation=True)
        )

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)
        expected_ret_say = [
            Word(word="hello", voice="test"),
            Word(word=",", voice="test", is_punctuation=True),
            Word(word=".", voice="test", is_punctuation=True),
        ]
        expected_ret_say.sort()
        expected_ret_unsay = words_from_strings(
            ["world", "duplicates", "duplicates!"], voice="test"
        )
        expected_ret_unsay.sort()

        assert not set(ret_say) ^ set(expected_ret_say)
        assert not set(ret_unsay) ^ set(expected_ret_unsay)

    def test_comp_sent(self, voice: SingleVoice) -> None:
        words = [
            Word(word="hello", voice="test"),
            Word(word=",", voice="test", is_punctuation=True),
            Word(word="world", voice="test"),
            Word(word=".", voice="test", is_punctuation=True),
            Word(word="vox", voice="test"),
            Word(word="can't", voice="test"),
            Word(word="say", voice="test"),
            Word(word="some", voice="test"),
            Word(word="of", voice="test"),
            Word(word="this", voice="test"),
            Word(word=".", voice="test", is_punctuation=True),
        ]

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)
        expected_ret_say = [
            Word(word="hello", voice="test"),
            Word(word=",", voice="test", is_punctuation=True),
            Word(word="vox", voice="test"),
            Word(word=".", voice="test", is_punctuation=True),
        ]
        expected_ret_unsay = [
            Word(word="world", voice="test"),
            Word(word="can't", voice="test"),
            Word(word="say", voice="test"),
            Word(word="some", voice="test"),
            Word(word="of", voice="test"),
            Word(word="this", voice="test"),
        ]
        assert not set(ret_say) ^ set(expected_ret_say)
        assert not set(ret_unsay) ^ set(expected_ret_unsay)

    def test_dup_punct(self, voice: SingleVoice) -> None:
        words = [
            Word(word="hello", voice="test"),
            Word(word=".", voice="test", is_punctuation=True),
            Word(word=".", voice="test", is_punctuation=True),
            Word(word=".", voice="test", is_punctuation=True),
            Word(word="world", voice="test"),
        ]

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)
        expected_ret_say = [
            Word(word="hello", voice="test"),
            Word(word=".", voice="test", is_punctuation=True),
        ]
        expected_ret_unsay = [
            Word(word="world", voice="test"),
        ]

        assert not set(ret_say) ^ set(expected_ret_say)
        assert not set(ret_unsay) ^ set(expected_ret_unsay)


class TestSentenceGeneration:
    def test_empty_sent(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("")

        assert ret.sentence == ""
        assert ret.sayable == []
        assert ret.unsayable == []
        assert ret.audio is None

        assert voice.get_generated_sentences_dict() == {}

    def test_unsayable_sent(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("whatthefuckdidyoujustsaytome")

        assert ret.sentence == ""
        assert ret.sayable == []
        assert ret.unsayable == [
            Word(word="whatthefuckdidyoujustsaytome", voice="test")
        ]
        assert ret.audio is None

        assert voice.get_generated_sentences_dict() == {}

    def test_sayable_sent(self, voice: SingleVoice) -> None:
        sentence = "hello, my name is vox"
        ret = voice.generate_audio(sentence)
        expected_sayable = [
            Word(",", voice="test", is_punctuation=True),
            Word("hello", voice="test"),
            Word("is", voice="test"),
            Word("my", voice="test"),
            Word("name", voice="test"),
            Word("vox", voice="test"),
        ]

        assert ret.sentence == "hello , my name is vox"
        assert ret.sayable == expected_sayable
        assert ret.unsayable == []
        assert ret.audio is not None
        assert ret.id == "1"

        assert voice.get_generated_sentences_dict() == {1: "hello , my name is vox"}

    def test_duplicate_sent(self, voice: SingleVoice) -> None:
        voice.generate_audio("hello hello")
        voice.generate_audio("hello hello")

        assert voice.get_generated_sentences_dict() == {1: "hello hello"}
        assert len(voice._get_generated_sentences_list()) == 1

    def test_duplicate_words(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello hello hello")

        assert ret.sentence == "hello hello hello"
        assert ret.sayable == [Word(word="hello", voice="test")]
        assert ret.unsayable == []
        assert ret.audio is not None

    def test_dup_punct(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello... hello")

        assert ret.sentence == "hello . . . hello"
        assert ret.sayable == [
            Word(word=".", voice="test", is_punctuation=True),
            Word(word="hello", voice="test"),
        ]
        assert ret.unsayable == []
        assert ret.audio is not None

    def test_multiple_sent(self, voice: SingleVoice) -> None:
        first_sentence = voice.generate_audio("hello hello")
        second_sentence = voice.generate_audio("vox hello")

        assert voice.get_generated_sentences_dict() == {
            1: "hello hello",
            2: "vox hello",
        }

        assert first_sentence.id == "1"
        assert second_sentence.id == "2"

    def test_dry_run(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello", dry_run=True)
        assert ret.audio is None

    def test_one_word(self, voice: SingleVoice) -> None:
        # We don't bother saving single words to the database
        voice.generate_audio("hello")

        assert voice.get_generated_sentences_dict() == {}

    def test_fetch_by_id(self, voice: SingleVoice) -> None:
        sentence = "hello, my name is vox"
        voice.generate_audio(sentence)

        sentence_from_id = voice.generate_audio_from_id(sentence_id="1")
        assert sentence_from_id.sentence == "hello , my name is vox"

        # IDs that don't exist should raise an exception
        with pytest.raises(SentenceIdNotFound):
            voice.generate_audio_from_id(sentence_id="2")


class TestModifiers:
    def test_basic_modifier(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|z1.0")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.NoOpModifier1(parameter=1.0)],
                is_punctuation=False,
            )
        ]

    def test_duplicates(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|z1.0,x3,z2,z2.1,x0.1")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[
                    # We should only get the first instance of a modifier
                    hlvox.voice.NoOpModifier2(parameter=3.0),
                    hlvox.voice.NoOpModifier1(parameter=1.0),
                ],
                is_punctuation=False,
            )
        ]

    def test_invalid_modifier(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|z1.0 my|k2")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.NoOpModifier1(parameter=1.0)],
                is_punctuation=False,
            ),
            # For now, we just ignore invalid modifiers
            Word(
                word="my",
                voice="test",
            ),
        ]

    def test_complicated(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello.|z1.0,x2 my, .|z0.1 name|z2 is, vox")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[
                    # Sorted alphabetically, which is not the same order it was written in
                    hlvox.voice.NoOpModifier2(parameter=2.0),
                    hlvox.voice.NoOpModifier1(parameter=1.0),
                ],
                is_punctuation=False,
            ),
            # This punctuation gets pulled out from the modified portion, it doesn't get modified
            Word(word=".", voice="test", is_punctuation=True),
            Word(
                word="my",
                voice="test",
            ),
            Word(
                word=",",
                voice="test",
                is_punctuation=True,
            ),
            # Again, punctuation doesn't get modified
            Word(
                word=".",
                voice="test",
                is_punctuation=True,
            ),
            Word(
                word="name",
                voice="test",
                modifiers=[
                    hlvox.voice.NoOpModifier1(parameter=2.0),
                ],
            ),
            Word(
                word="is",
                voice="test",
            ),
            Word(
                word=",",
                voice="test",
                is_punctuation=True,
            ),
            Word(
                word="vox",
                voice="test",
            ),
        ]
        assert ret.sentence == "hello|x2.0,z1.0 . my , . name|z2.0 is , vox"

    def test_equality(self) -> None:
        speed1 = hlvox.voice.SpeedChangeModifier(speed_parameter=1.0)
        speed2 = hlvox.voice.SpeedChangeModifier(speed_parameter=2.0)
        speed3 = hlvox.voice.SpeedChangeModifier(speed_parameter=1.0)

        assert speed1 != speed2
        assert speed1 == speed3

        pitch1 = hlvox.voice.PitchChangeModifier(pitch_parameter=1.0)
        pitch2 = hlvox.voice.PitchChangeModifier(pitch_parameter=2.0)
        pitch3 = hlvox.voice.PitchChangeModifier(pitch_parameter=1.0)

        assert pitch1 != pitch2
        assert pitch1 == pitch3

        acc1 = hlvox.voice.AcceleratorModifier(
            speed_increment=1,
            count=2,
            direction=hlvox.voice.AcceleratorDirection.FORWARD,
        )
        acc2 = hlvox.voice.AcceleratorModifier(
            speed_increment=1,
            count=2,
            direction=hlvox.voice.AcceleratorDirection.REVERSE,
        )
        acc3 = hlvox.voice.AcceleratorModifier(
            speed_increment=1,
            count=2,
            direction=hlvox.voice.AcceleratorDirection.FORWARD,
        )
        acc4 = hlvox.voice.AcceleratorModifier(
            speed_increment=1,
            count=3,
            direction=hlvox.voice.AcceleratorDirection.FORWARD,
        )
        acc5 = hlvox.voice.AcceleratorModifier(
            speed_increment=2,
            count=2,
            direction=hlvox.voice.AcceleratorDirection.FORWARD,
        )

        assert acc1 != acc2
        assert acc1 == acc3
        assert acc1 != acc4
        assert acc1 != acc5

    def test_accelerator(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|a2.2+100 hello|a2+10+f hello|a3+2+r")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[
                    hlvox.voice.AcceleratorModifier(
                        speed_increment=2.2,
                        count=100,
                        direction=hlvox.voice.AcceleratorDirection.FORWARD,
                    )
                ],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[
                    hlvox.voice.AcceleratorModifier(
                        speed_increment=2,
                        count=10,
                        direction=hlvox.voice.AcceleratorDirection.FORWARD,
                    )
                ],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[
                    hlvox.voice.AcceleratorModifier(
                        speed_increment=3,
                        count=2,
                        direction=hlvox.voice.AcceleratorDirection.REVERSE,
                    )
                ],
                is_punctuation=False,
            ),
        ]

    def test_accelerator_abnormal(self, voice: SingleVoice) -> None:
        with pytest.raises(hlvox.voice.ModifierArgumentsInvalid):
            voice.generate_audio("hello|a1.2+-100")

        with pytest.raises(hlvox.voice.ModifierArgumentsInvalid):
            voice.generate_audio("hello|a-1.2+100")

        with pytest.raises(hlvox.voice.ModifierFormatNotCorrect):
            voice.generate_audio("hello|a1.2+100+r+f")

        with pytest.raises(hlvox.voice.AcceleratorDirectionInvalid):
            voice.generate_audio("hello|a-1.2+100+x")

    def test_pitch_change(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|p1.0 hello|p-1.0 hello|p2")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.PitchChangeModifier(pitch_parameter=1.0)],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.PitchChangeModifier(pitch_parameter=-1.0)],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.PitchChangeModifier(pitch_parameter=2.0)],
                is_punctuation=False,
            ),
        ]

    def test_speed_change(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|s1.0 hello|s2")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.SpeedChangeModifier(speed_parameter=1.0)],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.SpeedChangeModifier(speed_parameter=2.0)],
                is_punctuation=False,
            ),
        ]

    def test_speed_change_abnormal(self, voice: SingleVoice) -> None:
        with pytest.raises(hlvox.voice.ModifierArgumentsInvalid):
            voice.generate_audio("hello|s-1.0 hello|s0.0")

    def test_volume_change(self, voice: SingleVoice) -> None:
        ret = voice.generate_audio("hello|v1.0 hello|v2")

        assert ret.sayable_sentence == [
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.VolumeChangeModifier(volume_parameter=1.0)],
                is_punctuation=False,
            ),
            Word(
                word="hello",
                voice="test",
                modifiers=[hlvox.voice.VolumeChangeModifier(volume_parameter=2.0)],
                is_punctuation=False,
            ),
        ]


class TestGetGeneratedSentences:
    def test_no_sentences(self, voice: SingleVoice) -> None:
        ret = voice.get_generated_sentences()

        assert ret == []

    def test_single_sentences(self, voice: SingleVoice) -> None:
        voice.generate_audio("hello hello")

        ret = voice.get_generated_sentences()

        assert ret == ["hello hello"]

    def test_multiple_sentences(self, voice: SingleVoice) -> None:
        voice.generate_audio("hello hello")
        voice.generate_audio("vox hello")

        ret = voice.get_generated_sentences()

        assert ret == ["hello hello", "vox hello"]


class TestGetGeneratedSentencesDict:
    def test_no_sentences(self, voice: SingleVoice) -> None:
        ret = voice.get_generated_sentences_dict()

        assert ret == {}

    def test_single_sentences(self, voice: SingleVoice) -> None:
        voice.generate_audio("hello hello")

        ret = voice.get_generated_sentences_dict()

        assert ret == {1: "hello hello"}

    def test_multiple_sentences(self, voice: SingleVoice) -> None:
        voice.generate_audio("hello hello")
        voice.generate_audio("vox hello")

        ret = voice.get_generated_sentences_dict()

        assert ret == {1: "hello hello", 2: "vox hello"}


@pytest.fixture(name="multi_voice")
def multi_voice_fixture(tmp_path: Path) -> Generator[MultiVoice, None, None]:
    norm_voice_dir = th.create_voice_files(tmp_path, normal_files, voice_name="norm")
    alph_voice_dir = th.create_voice_files(tmp_path, alph_files, voice_name="alph")
    alph_2_voice_dir = th.create_voice_files(tmp_path, alph_files, voice_name="alph2")
    norm_voice = SingleVoice(
        name="norm",
        path=norm_voice_dir,
        database=sqlalchemy.create_engine(f"sqlite:///{tmp_path}/norm.sqlite"),
    )
    alph_voice = SingleVoice(
        name="alph",
        path=alph_voice_dir,
        database=sqlalchemy.create_engine(f"sqlite:///{tmp_path}/alph.sqlite"),
    )
    alph_2_voice = SingleVoice(
        name="alph2",
        path=alph_2_voice_dir,
        database=sqlalchemy.create_engine(f"sqlite:///{tmp_path}/alph2.sqlite"),
    )
    voices = {
        "norm": norm_voice,
        "alph": alph_voice,
        "alph2": alph_2_voice,
    }
    multi_voice = MultiVoice(
        voices=voices,
        database=sqlalchemy.create_engine(f"sqlite:///{tmp_path}/multi.sqlite"),
    )
    yield multi_voice
    multi_voice.exit()
    for voice in voices.values():
        voice.exit()


EXPECTED_MULTI_VOICE_WORDS = [
    Word(word="hello", voice="norm"),
    Word(word="is", voice="norm"),
    Word(word="my", voice="norm"),
    Word(word="name", voice="norm"),
    Word(word="vox", voice="norm"),
    Word(word="a", voice="alph"),
    Word(word="b", voice="alph"),
    Word(word="c", voice="alph"),
    Word(word="a", voice="alph2"),
    Word(word="b", voice="alph2"),
    Word(word="c", voice="alph2"),
]

# Multi-voice tests


class TestMultiWords:
    def test_normal(self, multi_voice: MultiVoice) -> None:
        assert multi_voice.words == EXPECTED_MULTI_VOICE_WORDS


class TestMultiProcessSentence:
    # Just want to check that punctuation works as expected
    def test_normal(self, multi_voice: MultiVoice) -> None:
        words = [
            Word(word="norm:hello,"),
            Word(word=","),
            Word(word="norm:my"),
            Word(word="norm:name"),
            Word(word="alph:a,"),
            Word(word="alph2:b"),
            Word(word="."),
        ]
        exp_words = [
            Word(word="norm:hello"),
            Word(word=",", is_punctuation=True),
            Word(word=",", is_punctuation=True),
            Word(word="norm:my"),
            Word(word="norm:name"),
            Word(word="alph:a"),
            Word(word=",", is_punctuation=True),
            Word(word="alph2:b"),
            Word(word=".", is_punctuation=True),
        ]
        ret = multi_voice.process_sentence(words, voice=None)
        assert ret == exp_words


class TestMultiSayableUnsayable:
    def test_empty(self, multi_voice: MultiVoice) -> None:
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=[])
        assert not sayable
        assert not unsayable

    def test_all_sayable(self, multi_voice: MultiVoice) -> None:
        words = [
            Word(word="norm:hello"),
            Word(word="norm:my"),
            Word(word="alph:a"),
            Word(word="alph2:a"),
        ]
        exp_words = [
            Word(word="hello", voice="norm"),
            Word(word="my", voice="norm"),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph2"),
        ]
        exp_words.sort()
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        words.sort()
        assert sayable == exp_words
        assert not unsayable

    def test_some_unsayable(self, multi_voice: MultiVoice) -> None:
        words = words_from_strings(
            ["norm:nope", "alph:butter", "alph2:404", "norm:vox", "alph:c", "alph2:c"]
        )
        exp_unsayable = [
            Word(word="nope", voice="norm"),
            Word(word="butter", voice="alph"),
            Word(word="404", voice="alph2"),
        ]
        exp_sayable = [
            Word(word="vox", voice="norm"),
            Word(word="c", voice="alph"),
            Word(word="c", voice="alph2"),
        ]
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        exp_sayable.sort()
        exp_unsayable.sort()
        assert sayable == exp_sayable
        assert unsayable == exp_unsayable

    def test_punctuation(self, multi_voice: MultiVoice) -> None:
        words = [
            Word(word="norm:hello"),
            Word(word=",", is_punctuation=True),
            Word(word="alph:a", voice="alph"),
            Word(word=".", is_punctuation=True),
        ]
        exp_sayable = [
            Word(word="hello", voice="norm"),
            Word(word=",", voice="norm", is_punctuation=True),
            Word(word="a", voice="alph"),
            Word(word=".", voice="alph", is_punctuation=True),
        ]
        exp_sayable.sort()
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        assert sayable == exp_sayable
        assert not unsayable


class TestMultiWordVoiceAssignment:
    def test_empty(self, multi_voice: MultiVoice) -> None:
        assert not multi_voice._get_word_voice_assignment(words=[])

    def test_no_initial_voice(self, multi_voice: MultiVoice) -> None:
        words = words_from_strings(["hello", "norm:my"])

        with pytest.raises(hlvox.voice.NoVoiceSpecified):
            multi_voice._get_word_voice_assignment(words=words)

    def test_normal(self, multi_voice: MultiVoice) -> None:
        words = words_from_strings(
            ["norm:hello", "my", "alph:a", "a", "alph2:a", "norm:vox", "is", "alph:b"]
        )
        exp_words_and_voices = [
            Word(word="hello", voice="norm"),
            Word(word="my", voice="norm"),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph2"),
            Word(word="vox", voice="norm"),
            Word(word="is", voice="norm"),
            Word(word="b", voice="alph"),
        ]

        assert (
            multi_voice._get_word_voice_assignment(words=words) == exp_words_and_voices
        )

    def test_punctuation(self, multi_voice: MultiVoice) -> None:
        words = words_from_strings(
            [
                "norm:hello",
                ",",
                "alph:a",
                "a",
                ".",
                "alph2:a",
                "norm:vox",
                "is",
                "alph:b",
            ]
        )
        exp_words_and_voices = [
            Word(word="hello", voice="norm"),
            Word(word=",", voice="norm", is_punctuation=True),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph"),
            Word(word=".", voice="alph", is_punctuation=True),
            Word(word="a", voice="alph2"),
            Word(word="vox", voice="norm"),
            Word(word="is", voice="norm"),
            Word(word="b", voice="alph"),
        ]

        assert (
            multi_voice._get_word_voice_assignment(words=words) == exp_words_and_voices
        )


class TestMultiCombinedVoiceSentences:
    def test_empty(self, multi_voice: MultiVoice) -> None:
        assert not multi_voice.get_combined_voice_sentences(words=[])

    def test_single(self, multi_voice: MultiVoice) -> None:
        voices = multi_voice._voices
        words_and_voices = [Word(word="hello", voice="norm")]

        exp = [(voices["norm"], [Word(word="hello", voice="norm")])]
        assert multi_voice.get_combined_voice_sentences(words=words_and_voices) == exp

    def test_multiple(self, multi_voice: MultiVoice) -> None:
        voices = multi_voice._voices

        words_and_voices = [
            Word(word="hello", voice="norm"),
            Word(word="my", voice="norm"),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph"),
            Word(word="a", voice="alph2"),
            Word(word="vox", voice="norm"),
            Word(word="is", voice="norm"),
            Word(word="b", voice="alph"),
        ]

        exp = [
            (voices["norm"], words_from_strings(["hello", "my"], voice="norm")),
            (voices["alph"], words_from_strings(["a", "a"], voice="alph")),
            (voices["alph2"], words_from_strings(["a"], voice="alph2")),
            (voices["norm"], words_from_strings(["vox", "is"], voice="norm")),
            (voices["alph"], words_from_strings(["b"], voice="alph")),
        ]
        assert multi_voice.get_combined_voice_sentences(words=words_and_voices) == exp


class TestMultiGenerateAudio:
    def test_normal(self, multi_voice: MultiVoice) -> None:
        sentence = "norm:hello, my name is vox. alph:a b c"
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:hello , my name is vox . alph:a b c"
        assert ret.sentence == exp_sentence
        # Sentence should not be saved into the individual voice's database
        assert len(multi_voice._voices["norm"].get_generated_sentences()) == 0
        assert len(multi_voice._voices["alph"].get_generated_sentences()) == 0


class TestDifficultSentences:
    def test_ellipse(self, multi_voice: MultiVoice) -> None:
        sentence = "norm:hello norm:. norm:. norm:."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:hello . . ."
        assert ret.sentence == exp_sentence

        sentence = "norm:hello..."
        ret = multi_voice.generate_audio(sentence=sentence)
        exp_sentence = "norm:hello . . ."
        assert ret.sentence == exp_sentence

    def test_start_ellipse(self, multi_voice: MultiVoice) -> None:
        sentence = "norm:."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:."
        assert ret.sentence == exp_sentence

        sentence = "norm:. . ."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:. . ."
        assert ret.sentence == exp_sentence

        sentence = "norm:..."
        ret = multi_voice.generate_audio(sentence=sentence)
        assert ret.sentence == exp_sentence


class TestMultiModifiers:
    def test_normal(self, multi_voice: MultiVoice) -> None:
        sentence = "norm:hello|z1 norm:.|x2.0 norm:. norm:."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_words = [
            Word(
                word="hello",
                voice="norm",
                modifiers=[hlvox.voice.NoOpModifier1(parameter=1.0)],
                is_punctuation=False,
            ),
            Word(
                word=".",
                voice="norm",
                modifiers=[],
                is_punctuation=True,
            ),
            Word(
                word=".",
                voice="norm",
                modifiers=[],
                is_punctuation=True,
            ),
            Word(
                word=".",
                voice="norm",
                modifiers=[],
                is_punctuation=True,
            ),
        ]
        assert ret.sayable_sentence == exp_words

        exp_sentence = "norm:hello|z1.0 . . ."
        assert ret.sentence == exp_sentence



"""Defines a unique voice"""

import copy
import enum
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import (
    BinaryIO,
    Dict,
    List,
    Optional,
    Tuple,
    TypeAlias,
    Union,
    Protocol,
    Type,
)
from types import TracebackType

import librosa
import numpy as np
import soundfile as sf  # type: ignore[import-untyped]
import sqlalchemy

log = logging.getLogger(__name__)


@dataclass
class Audio:
    """Represents an audio file"""

    data: np.ndarray
    sample_rate: float

    @classmethod
    def from_file(cls, file_path: Path) -> "Audio":
        """Create audio segment from file

        Args:
            file_path (Path): Path to audio file

        Returns:
            Audio: audio segment
        """
        data, sample_rate = librosa.load(file_path, mono=True)
        return Audio(data=data, sample_rate=sample_rate)

    @classmethod
    def create_silence(cls, duration: float, sample_rate: float) -> "Audio":
        """Create silence audio segment

        Args:
            duration (float): duration of silence
            sample_rate (float): sample rate of silence

        Returns:
            Audio: silence audio segment
        """
        data = np.zeros(int(duration * sample_rate))
        return Audio(data=data, sample_rate=sample_rate)

    def get_sample_rate(self) -> float:
        """Get sample rate of audio segment

        Returns:
            float: sample rate
        """
        return self.sample_rate

    def set_sample_rate(self, sample_rate: float) -> None:
        """Set sample rate of audio segment

        Args:
            sample_rate (int): sample rate to set

        Returns:
            Audio: audio segment with new frame rate
        """
        resampled_data = librosa.resample(
            y=self.data,
            orig_sr=self.sample_rate,
            target_sr=sample_rate,
        )
        self.data = resampled_data
        self.sample_rate = sample_rate

    def get_length_seconds(self) -> float:
        """Get length of audio segment in seconds

        Returns:
            float: length in seconds
        """
        return librosa.get_duration(y=self.data, sr=self.sample_rate)

    def export(self, output: BinaryIO, audio_format: str) -> None:
        """Export audio segment to file

        Args:
            output (BinaryIO): file to export to
            audio_format (str): audio format for export
        """
        sf.write(
            file=output,
            data=self.data,
            samplerate=self.sample_rate,
            format=audio_format,
        )

    def __add__(self, other: "Audio") -> "Audio":
        """Add two audio segments together

        Args:
            other (Audio): audio segment to add

        Returns:
            Audio: combined audio segment
        """
        # Set sample rate to highest common denominator
        if self.get_sample_rate() > other.get_sample_rate():
            other.set_sample_rate(self.get_sample_rate())
        elif self.get_sample_rate() < other.get_sample_rate():
            self.set_sample_rate(other.get_sample_rate())

        combined_data: np.ndarray = np.concatenate((self.data, other.data))
        return Audio(data=combined_data, sample_rate=self.get_sample_rate())


class ModifierClass(Protocol):
    IDENTIFIER: str

    @classmethod
    def from_str(cls, string: str) -> "Modifier": ...


@dataclass
class Modifier:
    """Base class for sound modifiers"""

    IDENTIFIER = ""

    @classmethod
    def from_str(cls, string: str) -> "Modifier":  # pylint: disable=unused-argument
        """Create from string representation"""
        return Modifier()

    def as_str(self) -> str:
        """Convert to string representation"""
        return ""

    def modify_audio(self, audio: Audio) -> Audio:
        """Modify audio segment

        Args:
            audio (Audio): audio segment to modify

        Returns:
            Audio: modified audio segment
        """
        return audio

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError


@dataclass
class SpeedChangeModifier(Modifier):
    """Modify speed of audio without changing pitch"""

    IDENTIFIER = "s"

    def __init__(self, speed_parameter: float) -> None:
        if speed_parameter <= 0.0:
            raise ModifierArgumentsInvalid

        self.speed_parameter = float(speed_parameter)

    @classmethod
    def from_str(cls, string: str) -> "SpeedChangeModifier":
        speed_parameter_str = string.strip(SpeedChangeModifier.IDENTIFIER)
        return SpeedChangeModifier(speed_parameter=float(speed_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.speed_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        stretched_data = librosa.effects.time_stretch(
            y=audio.data, rate=self.speed_parameter
        )
        return Audio(data=stretched_data, sample_rate=audio.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SpeedChangeModifier):
            return False
        return self.speed_parameter == other.speed_parameter


@dataclass
class PitchChangeModifier(Modifier):
    """Modify pitch of audio without changing speed"""

    IDENTIFIER = "p"

    def __init__(self, pitch_parameter: float) -> None:
        self.pitch_parameter = float(pitch_parameter)

    @classmethod
    def from_str(cls, string: str) -> "PitchChangeModifier":
        pitch_parameter_str = string.strip(PitchChangeModifier.IDENTIFIER)
        return PitchChangeModifier(pitch_parameter=float(pitch_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.pitch_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        pitch_shifted_data = librosa.effects.pitch_shift(
            y=audio.data,
            sr=audio.sample_rate,
            n_steps=self.pitch_parameter,
        )
        return Audio(data=pitch_shifted_data, sample_rate=audio.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PitchChangeModifier):
            return False
        return self.pitch_parameter == other.pitch_parameter


class AcceleratorDirection(enum.Enum):
    """Direction for accelerator modifier"""

    FORWARD = enum.auto()
    REVERSE = enum.auto()

    @classmethod
    def from_str(cls, s: str) -> "AcceleratorDirection":
        """Convert from modifier string"""
        # TODO: this seems dumb
        if s == "f":
            return AcceleratorDirection.FORWARD
        if s == "r":
            return AcceleratorDirection.REVERSE
        raise AcceleratorDirectionInvalid

    def to_str(self) -> str:
        """Convert to modifier string"""
        # TODO: this seems dumb
        if self == AcceleratorDirection.FORWARD:
            return "f"
        if self == AcceleratorDirection.REVERSE:
            return "r"
        raise AcceleratorDirectionInvalid


@dataclass
class AcceleratorModifier(Modifier):
    """Accelerate a single word with modifiable pitch and speed curves"""

    IDENTIFIER = "a"

    def __init__(
        self, speed_increment: float, count: int, direction: AcceleratorDirection
    ) -> None:
        if speed_increment <= 0.0:
            raise ModifierArgumentsInvalid

        if count <= 1.0:
            raise ModifierArgumentsInvalid

        count = int(count)

        self.speed_increment = speed_increment
        self.count = count
        self.direction = direction

    # "hello|a2.0+100+r"
    # Where 2.0 is speed multiplier, 100 is the count, and r or f is forward or reverse. The last item is optional.
    @classmethod
    def from_str(cls, string: str) -> "AcceleratorModifier":
        parameter_str = string.strip(cls.IDENTIFIER)
        components = parameter_str.split("+")
        if len(components) < 2 or len(components) > 3:
            raise ModifierFormatNotCorrect
        speed_increment = float(components[0])
        count = int(components[1])
        direction = AcceleratorDirection.FORWARD
        if len(components) == 3:
            direction = AcceleratorDirection.from_str(components[2])

        return AcceleratorModifier(
            speed_increment=speed_increment, count=count, direction=direction
        )

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.speed_increment}+{self.count}+{self.direction.to_str()}"

    def modify_audio(self, audio: Audio) -> Audio:
        segment: Optional[Audio] = None
        increments: Union[range, reversed]

        if self.direction == AcceleratorDirection.REVERSE:
            increments = reversed(range(0, self.count))
        else:
            increments = range(0, self.count)

        for i in increments:
            if i == 0.0:
                speed_increment = 1.0
            else:
                speed_increment = i * self.speed_increment

            speed_modifier = SpeedChangeModifier(speed_parameter=speed_increment)

            modified_audio = copy.deepcopy(audio)
            if speed_modifier != 1.0:
                modified_audio = speed_modifier.modify_audio(audio=audio)
            else:
                modified_audio = copy.deepcopy(audio)

            if segment is None:
                segment = modified_audio
            else:
                segment += modified_audio

        if segment is None:
            raise ModifierArgumentsInvalid("No segments were created")

        return Audio(data=segment.data, sample_rate=segment.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AcceleratorModifier):
            return False
        return (
            self.speed_increment == other.speed_increment
            and self.count == other.count
            and self.direction == other.direction
        )


@dataclass
class VolumeChangeModifier(Modifier):
    """Modify volume of audio"""

    IDENTIFIER = "v"

    def __init__(self, volume_parameter: float) -> None:
        self.volume_parameter = float(volume_parameter)

    @classmethod
    def from_str(cls, string: str) -> "VolumeChangeModifier":
        volume_parameter_str = string.strip(VolumeChangeModifier.IDENTIFIER)
        return VolumeChangeModifier(volume_parameter=float(volume_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.volume_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        return Audio(
            data=audio.data * self.volume_parameter, sample_rate=audio.sample_rate
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, VolumeChangeModifier):
            return False
        return self.volume_parameter == other.volume_parameter


@dataclass
class NoOpModifier1(Modifier):
    """No-op modifier for testing"""

    IDENTIFIER = "z"

    def __init__(self, parameter: float) -> None:
        self.parameter = float(parameter)

    @classmethod
    def from_str(cls, string: str) -> "NoOpModifier1":
        parameter = string.strip(NoOpModifier1.IDENTIFIER)
        return NoOpModifier1(parameter=float(parameter))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.parameter}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoOpModifier1):
            return False
        return self.parameter == other.parameter


@dataclass
class NoOpModifier2(Modifier):
    """No-op modifier for testing"""

    IDENTIFIER = "x"

    def __init__(self, parameter: float) -> None:
        self.parameter = float(parameter)

    @classmethod
    def from_str(cls, string: str) -> "NoOpModifier2":
        parameter = string.strip(NoOpModifier2.IDENTIFIER)
        return NoOpModifier2(parameter=float(parameter))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.parameter}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoOpModifier2):
            return False
        return self.parameter == other.parameter


MODIFIERS: Dict[str, Type[ModifierClass]] = {
    SpeedChangeModifier.IDENTIFIER: SpeedChangeModifier,
    PitchChangeModifier.IDENTIFIER: PitchChangeModifier,
    AcceleratorModifier.IDENTIFIER: AcceleratorModifier,
    NoOpModifier1.IDENTIFIER: NoOpModifier1,
    NoOpModifier2.IDENTIFIER: NoOpModifier2,
    VolumeChangeModifier.IDENTIFIER: VolumeChangeModifier,
}


Modifiers: TypeAlias = List[Modifier]


class Word:
    """Represents a word (or punctuation)"""

    def __init__(
        self,
        word: str,
        voice: Optional[str] = None,
        is_punctuation: bool = False,
        modifiers: Optional[Modifiers] = None,
    ) -> None:
        self.word = word
        self.voice = voice
        self.modifiers = modifiers if modifiers else []
        self.modifiers.sort(key=lambda m: m.IDENTIFIER)
        self.is_punctuation = is_punctuation

    def as_str(self, with_voice: bool = False) -> str:
        """Convert to string representation

        Args:
            with_voice (bool, optional): Include voice string. Defaults to False.

        Returns:
            str: string representation
        """
        voice_string = f"{self.voice}:" if (with_voice and self.voice) else ""
        modifiers_string = ""
        if self.modifiers:
            modifier_strings = ",".join(
                [modifier.as_str() for modifier in self.modifiers]
            )
            modifiers_string = f"|{modifier_strings}"
        return f"{voice_string}{self.word}{modifiers_string}"

    def __str__(self) -> str:
        return self.as_str(with_voice=True)

    def __repr__(self) -> str:
        return self.as_str(with_voice=True)

    def without_modifiers(self) -> "Word":
        """Get a copy of the word without modifiers

        Returns:
            Word: word without modifiers
        """
        return Word(
            word=self.word, voice=self.voice, is_punctuation=self.is_punctuation
        )

    # TODO: sorting does not take into account modifiers
    def __lt__(self, other: "Word") -> bool:
        return self.word < other.word

    def __gt__(self, other: "Word") -> bool:
        return self.word > other.word

    def __le__(self, other: "Word") -> bool:
        return self.word <= other.word

    def __ge__(self, other: "Word") -> bool:
        return self.word >= other.word

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Word):
            return False
        return (
            self.word == other.word
            and self.modifiers == other.modifiers
            and self.is_punctuation == other.is_punctuation
            and self.voice == other.voice
        )

    # TODO: test that this matches __eq__ behavior
    def __hash__(self) -> int:
        return hash(repr(self))


# How much delay should be added in place of punctuation (in milliseconds)
PUNCTUATION_TIMING_SECONDS = {
    ",": 0.250,
    ".": 0.500,
}

DB_NAME = "db.json"


class NoWordsFound(Exception):
    """Raised when a voice has no words"""


class DuplicateWords(Exception):
    """Raised when a voice has duplicate words"""


class InconsistentAudioFormats(Exception):
    """Raised when words have inconsistent audio formats"""


class NoAudioFormatFound(Exception):
    """Raised when no audio format can be found"""


class FailedToSplit(Exception):
    """Raised when a sentence cannot be split"""


class NoVoiceSpecified(Exception):
    """Raised when no voice is specified"""


class NoDatabaseSpecified(Exception):
    """No database connection was specified during init"""


class ModifierSyntaxError(Exception):
    """Raised when there is a problem with the modifier syntax"""


class SentenceIdNotFound(Exception):
    """Could not find a previously generated sentence with specified ID"""


class ModifierFormatNotCorrect(Exception):
    """Modifiers not provided in the correct format"""


class AcceleratorDirectionInvalid(Exception):
    """Invalid accelerator direction"""


class ModifierArgumentsInvalid(Exception):
    """Invalid modifier arguments and/or format"""


@dataclass
class Sentence:
    """Represents a sentence and it's parts"""

    sentence: str
    sayable: List[Word]
    unsayable: List[Word]
    sayable_sentence: List[Word]
    audio: Optional[Audio] = None
    id: Optional[str] = None


@dataclass
class DatabaseConnection:
    """Stores info related to database connection"""

    engine: sqlalchemy.engine.Engine
    metadata: sqlalchemy.MetaData
    sentence_table: sqlalchemy.Table


class Voice:
    """Base class for Voice-like interfaces.
    Intended to involve generation of audio
    files from some source (files, web, etc).
    """

    def __init__(
        self,
        name: str,
        database: Optional[sqlalchemy.engine.Engine],
    ) -> None:
        self.name = name

        self._db: Optional[DatabaseConnection] = None
        if database is not None:
            metadata = sqlalchemy.MetaData()
            sentence_table = sqlalchemy.Table(
                self.name,
                metadata,
                sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column(
                    "sentence", sqlalchemy.String, unique=True, nullable=False
                ),
            )
            metadata.create_all(database)
            self._db = DatabaseConnection(
                metadata=metadata,
                sentence_table=sentence_table,
                engine=database,
            )
        self.words: List[Word] = []
        self.categories: Dict[str, List[str]] = {}

    def __enter__(self) -> "Voice":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.exit()

    def exit(self) -> None:
        """Clean up and close voice"""
        if self._db is not None:
            self._db.engine.dispose()

    def _insert_sentence_into_db(self, sentence: str) -> int:
        if self._db is None:
            raise NoDatabaseSpecified
        ins = self._db.sentence_table.insert().values(sentence=sentence)
        with self._db.engine.begin() as conn:
            result = conn.execute(ins)
        return int(result.inserted_primary_key[0])

    def _sentence_exists(self, sentence: str) -> bool:
        if self._db is None:
            raise NoDatabaseSpecified

        sel = self._db.sentence_table.select().where(
            self._db.sentence_table.c.sentence == sentence
        )
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
            # TODO: there should be a way to use `.Count()` here
        return bool(result.all())

    def _get_generated_sentences_list(self) -> list[str]:
        if self._db is None:
            raise NoDatabaseSpecified
        sel = self._db.sentence_table.select()
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
        return [r.sentence for r in result.all()]

    def _get_generated_sentences_dict(self) -> dict[int, str]:
        if self._db is None:
            raise NoDatabaseSpecified
        sel = self._db.sentence_table.select()
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
        return {r.id: r.sentence for r in result.all()}

    def _get_generated_sentence_from_id(self, sentence_id: str) -> Optional[str]:
        if self._db is None:
            raise NoDatabaseSpecified
        sel = self._db.sentence_table.select().where(
            # One day the ID might be a UUID string or something along those lines,
            # so we stay flexible on inputs and just convert to an int for now.
            self._db.sentence_table.c.id == int(sentence_id)
        )
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
        sentences = result.all()
        if not sentences:
            return None
        return str(sentences[0].sentence)

    def _get_sentence_info(self, words: List[Word]) -> Sentence:
        """Get basic sentence info for a given
        split sentence.

        Args:
            words (List[Word]): Words of sentence split into array

        Returns:
            Sentence: Sentence info
        """
        sayable_words, unsayable_worlds = self.get_sayable_unsayable(words)
        sayable_sent_arr = self._get_sayable_sentence_arr(words, sayable_words)
        sayable_sent_str = self._create_sentence_string(sayable_sent_arr)

        return Sentence(
            sentence=sayable_sent_str,
            sayable=sayable_words,
            unsayable=unsayable_worlds,
            sayable_sentence=sayable_sent_arr,
            audio=None,
            id=None,
        )

    def generate_audio_from_array(
        self, words: List[Word], dry_run: bool = False, save_to_db: bool = True
    ) -> Sentence:
        """Generates audio segment from sentence array.

        Args:
            words (List[str]): Words to try and turn into audio segment.
            dry_run (bool, optional): Skip actual segment generation. Defaults to False.

        Returns:
            Sentence: Sentence with audio segment.
        """
        sentence_info = self._get_sentence_info(words=words)

        if dry_run:
            return sentence_info

        log.debug("Generating %s", sentence_info.sentence)

        # Only create sentence if there are words to put in it
        if len(sentence_info.sayable) == 0:
            log.warning(
                "Can't say any words in %s, not generating", sentence_info.sentence
            )
            return sentence_info

        # Only bother inserting a sentence into the database if there is more than one word in it
        # TODO: test save_to_db
        if self._db and save_to_db and len(words) > 1:
            if not self._sentence_exists(sentence=sentence_info.sentence):
                sentence_id = self._insert_sentence_into_db(
                    sentence=sentence_info.sentence
                )
                sentence_info.id = str(sentence_id)
        words_audio = self._create_audio_segments(sentence_info.sayable_sentence)
        sentence_info.audio = self.assemble_audio_segments(words_audio)

        return sentence_info

    def _create_audio_segments(
        self,
        word_array: List[Word],  # pylint: disable=unused-argument
    ) -> List[Audio]:
        """Create audio segments for each entry in an array of words.

        Args:
            word_array (List[str]): Words to turn into audio segments.

        Returns:
            List[Audio]: Audio segments.
        """
        return []

    def generate_audio(self, sentence: str, dry_run: bool = False) -> Sentence:
        """Generates audio from the given sentence

        Args:
            sentence (string): Sentence string to be generated
            dry_run (bool, optional): Don't generate audio. Defaults to False.
        Returns:
            Sentence: Information about generated sentence.
        """
        log.info("Asked to generate %s", sentence)
        split_sentence = self._split_sentence(sentence)
        proc_sentence = self.process_sentence(split_sentence, voice=self.name)
        return self.generate_audio_from_array(
            words=proc_sentence,
            dry_run=dry_run,
        )

    def generate_audio_from_id(
        self, sentence_id: str, dry_run: bool = False
    ) -> Sentence:
        """Generate audio from an existing sentence, fetched by its ID

        Args:
            sentence_id (str): Existing sentence ID
            dry_run (bool, optional): Don't generate audio. Defaults to False.

        Raises:
            SentenceIdNotFound: Existing sentence ID not found in database

        Returns:
            Sentence: Information about generated sentence.
        """
        log.info("Fetching audio for ID %s", sentence_id)
        sentence = self._get_generated_sentence_from_id(sentence_id)
        if not sentence:
            raise SentenceIdNotFound
        return self.generate_audio(sentence=sentence, dry_run=dry_run)

    @staticmethod
    def _split_sentence(sentence: str) -> List[Word]:
        return [Word(word=word) for word in sentence.lower().rstrip().split(" ")]

    @staticmethod
    def _extract_modifiers(words: List[Word]) -> List[Word]:
        processed_words: list[Word] = []
        for word in words:
            if "|" not in word.word:
                processed_words.append(word)
                continue
            (word_string, _, modifiers_string) = word.word.rpartition("|")
            modifiers_strings = modifiers_string.split(",")

            # Dict so we can dedupe
            modifiers: dict[str, Modifier] = {}
            for modifier_string in modifiers_strings:
                modifier_class = MODIFIERS.get(modifier_string[0])
                # TODO: not bubbling up the invalid modifier doesn't seem right
                if modifier_class:
                    modifier = modifier_class.from_str(modifier_string)
                    if modifier.IDENTIFIER in modifiers:
                        continue
                    modifiers[modifier.IDENTIFIER] = modifier

            word.word = word_string
            word.modifiers = list(modifiers.values())
            processed_words.append(word)
        return processed_words

    @staticmethod
    def process_sentence(split_sent: List[Word], voice: Optional[str]) -> List[Word]:
        """
        Takes a normally formatted sentence and breaks it into base elements

        Args:
            split_sent (List[str]): words in sentence

        Returns:
            List[Word]: array of elements in sentence
        """
        # TODO: This could use some rethinking. Should be easier to just always break punctuation marks
        # into their own elements, rather than selectively only dealing with trailing ones.
        log.info("Processing sentence '%s'", split_sent)

        # First pass for modifiers
        split_sent = Voice._extract_modifiers(words=split_sent)

        # Pull out punctuation
        reduced_sent: list[Word] = []
        for item in split_sent:
            word_string = item.word
            # find first punctuation mark, if any
            first_punct: Optional[str] = None
            try:
                first_punct = next(
                    (
                        punct
                        for punct in PUNCTUATION_TIMING_SECONDS
                        if punct in word_string
                    )
                )
            except StopIteration:
                pass

            if first_punct:
                # Get its index
                first_punct_ind = word_string.find(first_punct)

                # Special case: If this is a multi voice sentence,
                # we don't want to rip the voice definition out of a singe-punctuation
                # mark. IE vox:.
                # TODO: This is a bit hacky. Would be great if this method doesn't
                # have to know about multi-voice syntax.
                if first_punct_ind >= 2 and word_string[first_punct_ind - 1] == ":":
                    reduced_sent.append(
                        Word(
                            word=word_string[: first_punct_ind + 1],
                            voice=voice,
                            modifiers=item.modifiers,
                        )
                    )
                    if len(word_string) >= first_punct_ind:
                        first_punct_ind += 1
                else:
                    # Add everything before punct (the word, if any)
                    if word_string[:first_punct_ind]:
                        reduced_sent.append(
                            Word(
                                word=word_string[:first_punct_ind],
                                voice=voice,
                                modifiers=item.modifiers,
                            )
                        )

                # Add all the punctuation if its actually punctuation
                # TODO: Figure out if I want to deal with types like ".hello" throwing out all the characters after the period.
                for punct in word_string[first_punct_ind:]:
                    if punct in PUNCTUATION_TIMING_SECONDS:
                        reduced_sent.append(
                            Word(word=punct, voice=voice, is_punctuation=True)
                        )

            else:
                # TODO: copying from a Word to a Word like this is ugly
                reduced_sent.append(
                    Word(word=word_string, voice=voice, modifiers=item.modifiers)
                )

        # Clean blanks from reduced_sent
        reduced_sent = [value for value in reduced_sent if value.word != ""]

        log.info("Sentence processed: '%s'", reduced_sent)
        return reduced_sent

    def get_sayable_unsayable(self, words: List[Word]) -> Tuple[List[Word], List[Word]]:
        """Get words that are sayable or unsayable
        from a list of words.

        Args:
            words (List[Word]): Words to check.

        Returns:
            Tuple[List[Word], List[Word]]: Sayable and unsayable words.
        """
        # TODO: This shouldn't need two separate processings of the same sentence. Sets, people. Sets!
        sayable_words_set = set(self.words)
        sayable_words_set.update(
            [
                Word(word=punct, voice=self.name, is_punctuation=True)
                for punct in PUNCTUATION_TIMING_SECONDS
            ]
        )

        no_modifier_words = [word.without_modifiers() for word in words]
        words_set = set((dict.fromkeys(no_modifier_words)))  # removes duplicates

        unsayable_set = words_set - sayable_words_set
        sayable_set = words_set - unsayable_set
        unsayable = list(unsayable_set)
        unsayable.sort()
        sayable = list(sayable_set)
        sayable.sort()
        return sayable, unsayable

    def _get_sayable_sentence_arr(
        self, words: List[Word], sayable_words: List[Word]
    ) -> List[Word]:
        """Removes words from sentence array that are not sayable.

        Args:
            words (List[Word]): Array of words in sentence, in order.
            sayable_words (List[Word]): Words from sentence that can actually be said.

        Returns:
            List[Word]: Words in sentence that are sayable, in order.
        """
        # TODO: This is just a simple set operation. Function probably isn't needed. At least change to using a set.
        return [
            word
            for word in words
            if Word(
                word=word.word, voice=word.voice, is_punctuation=word.is_punctuation
            )
            in sayable_words
        ]

    def _create_sentence_string(self, words: List[Word]) -> str:
        """Joins sentence array into a string.

        Args:
            words (List[str]): Words in sentence, in order.

        Returns:
            str: Sentence string.
        """
        if len(words) == 1:
            return words[0].as_str(with_voice=False)
        return " ".join([word.as_str(with_voice=False) for word in words])

    def get_generated_sentences(self) -> List[str]:
        """Gets the previously generated sentence strings

        Returns:
            List[str]: List of sentence strings generated previously
        """
        return self._get_generated_sentences_list()

    def get_generated_sentences_dict(self) -> Dict[int, str]:
        """Gets the previously generated sentence strings
        along with their corresponding ID in the database

        Returns:
            Dict[int, str]: Dict of sentence and id pairs
        """
        return self._get_generated_sentences_dict()

    @staticmethod
    def assemble_audio_segments(segments: List[Audio]) -> Audio:
        """Assemble audio segments into one audio segment.

        Args:
            segments (List[Audio]): Segments to assemble.

        Returns:
            Audio: Assembled audio segment.
        """
        # We set all audio segments to the lowest common frame rate
        # to avoid some really ugly artifacting when a low frame rate
        # clip is appended to a high frame rate one.
        frame_rates = [word.get_sample_rate() for word in segments]
        frame_rate = min(frame_rates)

        sentence_audio = segments.pop(0)
        if sentence_audio.get_sample_rate() != frame_rate:
            sentence_audio.set_sample_rate(frame_rate)
        for word_audio in segments:
            if word_audio.get_sample_rate() != frame_rate:
                word_audio.set_sample_rate(frame_rate)
            sentence_audio = sentence_audio + word_audio

        return sentence_audio



"""Voice class that uses other voices to assemble
multi-voice sentences.
"""

import logging
from typing import Dict, Optional, Tuple
import sqlalchemy

from .single_voice import SingleVoice
from .voice import (
    PUNCTUATION_TIMING_SECONDS,
    Audio,
    FailedToSplit,
    NoVoiceSpecified,
    Sentence,
    Voice,
    Word,
)

log = logging.getLogger(__name__)


class MultiVoice(Voice):
    """Voice class that uses other voices to assemble
    multi-voice sentences.

    Example: vox:hello hev:there
    Generates a sentence with one word from a voice
    called "vox" and another from a voice called "hev."
    """

    def __init__(
        self, voices: dict[str, SingleVoice], database: sqlalchemy.engine.Engine | None
    ):
        """
        Args:
            voices (Dict[str, SingleVoice]): Voices to use to assemble sentences.
            database (Optional[sqlalchemy.engine.Engine]): Database connection information.
                If none provided, no database will be used and no data will persist.
        """
        super().__init__(name="multi", database=database)
        self._voices = voices

        self.words = self._get_words(voices)

    def _get_words(self, voices: Dict[str, SingleVoice]) -> list[Word]:
        words = []
        for _, voice in voices.items():
            voice_words = voice.words.copy()
            words.extend(voice_words)
        return words

    def _get_sentence_info(self, words: list[Word]) -> Sentence:
        # TODO: There is a good amount of double-processing going on here
        words_and_voices = self._get_word_voice_assignment(words)
        sayable_words, unsayable_words = self.get_sayable_unsayable(words)
        sayable_sent_arr = [
            word_voice
            for word_voice in words_and_voices
            if word_voice.without_modifiers() in sayable_words
        ]
        combined_voice_sentences = self.get_combined_voice_sentences(words_and_voices)
        sentence_arr = []
        for voice, sentence_words in combined_voice_sentences:
            voice_sentence_segment = (
                f"{voice.name}:{' '.join([word.as_str() for word in sentence_words])}"
            )
            sentence_arr.append(voice_sentence_segment)

        sayable_sent_str = " ".join(sentence_arr)

        return Sentence(
            sentence=sayable_sent_str,
            sayable=sayable_words,
            unsayable=unsayable_words,
            sayable_sentence=sayable_sent_arr,
            audio=None,
        )

    def get_sayable_unsayable(self, words: list[Word]) -> Tuple[list[Word], list[Word]]:
        sayable = []
        unsayable = []
        words_and_voices = self._get_word_voice_assignment(words=words)
        combined_voice_sentences = self.get_combined_voice_sentences(words_and_voices)
        for voice, sentence_words in combined_voice_sentences:
            voice_sayable, voice_unsayable = voice.get_sayable_unsayable(sentence_words)
            sayable.extend(voice_sayable)
            unsayable.extend(voice_unsayable)
        sayable.sort()
        unsayable.sort()
        return sayable, unsayable

    def _create_audio_segments(self, word_array: list[Word]) -> list[Audio]:
        combined_voice_sentences = self.get_combined_voice_sentences(word_array)
        return self.get_combined_audio(
            voice_sentences=combined_voice_sentences,
        )

    def _get_word_voice_assignment(self, words: list[Word]) -> list[Word]:
        """Determines voice for each word in a list separated
        from a raw sentence. Only the first word must have a voice
        assignment, further assignments are inferred.

        Example: vox:hello there hev:doctor freeman
        The first two words are assigned to vox, second two to hev

        Args:
            words (List[Word]): Words to determine voice assignment of

        Raises:
            FailedToSplit: Raised if unable to split a word/voice assignment.
            NoVoiceSpecified: Raised if initial voice cannot be determined.

        Returns:
            List[Word]: word:voice assignments
        """
        words_and_voices = []

        current_voice: Optional[SingleVoice] = None
        for word_maybe_voice in words:
            word_split = word_maybe_voice.word.split(":")
            word: Optional[Word] = None
            word_str: Optional[str] = None
            if len(word_split) == 2:
                voice_name = word_split[0]
                if voice_name not in self._voices:
                    raise NoVoiceSpecified(f"Voice {voice_name} not found")
                current_voice = self._voices[voice_name]
                word_str = word_split[1]
            elif len(word_split) == 1:
                word_str = word_split[0]

            if not word_str:
                raise FailedToSplit
            if not current_voice:
                raise NoVoiceSpecified("No voice specified for word")

            is_punctuation = word_str in PUNCTUATION_TIMING_SECONDS
            modifiers = word_maybe_voice.modifiers if not is_punctuation else []
            word = Word(
                word=word_str,
                voice=current_voice.name,
                modifiers=modifiers,
                is_punctuation=is_punctuation,
            )

            words_and_voices.append(word)

        return words_and_voices

    def get_combined_voice_sentences(
        self, words: list[Word]
    ) -> list[Tuple[SingleVoice, list[Word]]]:
        """Turns individual word:voice assignments into
        combined sentences for each word in sequence:

        Example: vox:hello vox:there hev:doctor hev:freeman vox:boop
        Returns vox:[hello, there] hev:[doctor freeman] vox:[boop]

        Args:
            words (List[Word]): Word:voice assignments

        Returns:
            List[Tuple[SingleVoice, List[Word]]]: Voice:sentence assignments
        """
        current_voice: Optional[SingleVoice] = None
        current_voice_sentence: list[Word] = []
        voice_sentences = []
        for word in words:
            if not current_voice:
                if word.voice not in self._voices:
                    raise NoVoiceSpecified(f"Voice {word.voice} not found")
                current_voice = self._voices[word.voice]
            if word.voice == current_voice.name:
                current_voice_sentence.append(word)
            else:
                voice_sentences.append((current_voice, current_voice_sentence))
                if word.voice not in self._voices:
                    raise NoVoiceSpecified(f"Voice {word.voice} not found")
                current_voice = self._voices[word.voice]
                current_voice_sentence = [word]
        if current_voice and current_voice_sentence:
            voice_sentences.append((current_voice, current_voice_sentence))
        return voice_sentences

    def get_combined_audio(
        self, voice_sentences: list[Tuple[SingleVoice, list[Word]]]
    ) -> list[Audio]:
        """Generates audio segments for each voice sentence

        Args:
            voice_sentences (List[Tuple[SingleVoice, List[str]]]): Voice:sentence assignments

        Returns:
            List[Audio]: List of generated audio segments
        """
        audio_segments = []
        for voice, words in voice_sentences:
            sentence = voice.generate_audio_from_array(words, save_to_db=False)
            if sentence.audio:
                audio_segments.append(sentence.audio)
        sample_rate = max(audio.get_sample_rate() for audio in audio_segments)
        for audio in audio_segments:
            if audio.get_sample_rate() != sample_rate:
                audio.set_sample_rate(sample_rate)
        return audio_segments
